//=============================================================================
// Filename: lichtschranke.c
//=============================================================================
// Author: 
// Company:
// Revision: 1.0
// Date:   11/10/2006
// letzte nderung: 11.10.2006
//=============================================================================
// LichtschrankeV3 Beschreibung:
// 
//
//
//=============================================================================
//
// Taktfrequenz: 4,9152 MHz Quarz 
// PIC: PIC18F1320
//
//
// Wichtig: Damit es keinen einstellungsbedingten Resourcen-Mangel gibt:
// MPLAB IDE: Projekt/Build Options/Projekt: MPLAB C18, Categories: 
// Memory Model: 
// Data model:  small data model (Access RAM only)
// Stack model: single bank model
//=============================================================================

#pragma config WDT = OFF
#pragma config OSC = XT
#pragma config PWRT = ON
#pragma config LVP = OFF
#pragma config BOR = ON
#pragma config BORV = 27
#pragma config FSCM = ON
#pragma config IESO = ON
#pragma config MCLRE = ON
#pragma config CP0 = OFF
#pragma config CP1 = OFF
//----------------------------------------------------------------------------

#include <p18F1320.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

//----------------------------------------------------------------------------
// fr den C18-Compiler gilt: int = short = 16 Bit !
//----------------------------------------------------------------------------



#define	CLK		4915200
//#define	CLK		3686400	//Quarz Frequenz
#define	BAUD	9600
#define	TRUE	1
#define	FALSE	0
#define	LS1		1	//fr Lichtschranke 1
#define	LS2		2	//fr Lichtschranke 2 
#define	MODE12	1	//Betrieb mit 2 Lichtschranken
#define	MODE1	0	//Betrieb mit Lichtschranke 1 ausschliesslich

#define	CMD_STOP	0x1c	//ESC = Stop per UART
#define	CMD_ACK		'h'
#define	ACK			'H'		//Hallo, fr RS232 Test

#define	RESET	PORTBbits.RB3
#define	MODE	PORTAbits.RA0
#define	LED		PORTAbits.RA1
#define	SIG1	PORTAbits.RA2
#define	SIG2	PORTAbits.RA3

// Funktions - Prototypen:
void 	main (void);
void 	high_isr( void );
void 	HWInit( void );
void 	SendData( void );
void 	SendChar( void );
void 	delay( unsigned char );
void	lichtsch( int );
void 	FReset( void );
void	SendByte( char );


// I/O - Pin belegung des PIC18:
//Quarz Primrer takt an OSC1,2
//MCLR als Reset
// RB7
// RB6
// RB5
// RB4 IN, UART RXD
// RB3 IN, RESET
// RB2 IN, LS2, ber INT2
// RB1 OUT, UART TXD
// RB0 IN, LS1, ber INT0
//
// RA4
// RA3 OUT, SIG1
// RA2 OUT, SIG2
// RA1 OUT, LED
// RA0 IN, MODE12
//
// Verwendete Funktionen:
// UART
// Timer 1 fr Zeitmessung, Int alle 1ms
// Timer 3 fr Funktion delay
// keine analoge Funktion
//

union BI{			//union fr byte-Int-Umwandlung
	unsigned int i;
	unsigned char b[2];
};

//----------------------------------------------------------------
// Globale Variablen:
#define	TXDLEN	32
int		dummyg, countf, ctr, mode, dataf, txdctr, txdidx;
volatile int ct10;
int		txdready;
char	txdbuf[TXDLEN];
union 	BI tmr1;

//----------------------------------------------------------------
// Interrupts:
// High priority interrupt vector
#pragma interrupt high_isr
void high_isr( void )
{	
	if( INTCONbits.INT0IF == 1 )
	{
		INTCONbits.INT0IE = 0;
		INTCONbits.INT0IF = 0;
		lichtsch( LS1 );
		INTCONbits.INT0IF = 0;			//LS1 - Interrupt
		//INTCONbits.INT0IE = 1;
	}
	else if( INTCON3bits.INT2IF == 1 )	//LS2 - Interrupt
	{
		INTCON3bits.INT2IE = 0;
		INTCON3bits.INT2IF = 0;
		lichtsch( LS2 );
		INTCON3bits.INT2IF = 0;
		//INTCON3bits.INT2IE = 1;
	}
	else if( PIR1bits.TMR1IF == 1 )		//Timer 1 Int
	{
		PIR1bits.TMR1IF = 0;
		TMR1L = tmr1.b[0];			//Lichtschranken-Funktion
		TMR1H = tmr1.b[1];
		if( countf != 0 )
		{	
			ctr++;
			ct10--;
			if( ct10 == 0 )
			{
				SendData();
				ct10 = 10;
			}
		}
	}
	else if( PIR1bits.RCIF == 1 )		//RX Interrupt
	{
		dummyg = RCREG;
		if( dummyg == CMD_STOP )
			FReset();
		if( dummyg == CMD_ACK )
			SendByte( ACK );
	}
    else if( PIR1bits.TXIF == 1 )		//TX Interrupt
    	SendChar();
}
// Interrupt - Vektor:
#pragma code high_vector = 0x08
void interrupt_at_high_vector( void )
{
  _asm
    goto high_isr //jump to interrupt routine
  _endasm
}
#pragma code
//
// Dieser hier wird nicht verwendet - also nur Dekoration:
#pragma code low_vector = 0x18
void	interrupt_at_low_vector( void )
{
	if( PIR1bits.RCIF == 1 )
		dummyg = RCREG;
	_asm
    goto high_isr //jump to interrupt routine
  _endasm
}
#pragma code

//--------------------------------------------------------------------------
// Main routine

void main ()
{
	int i;
	
	tmr1.i = 65536-CLK/400;	//fr Timer 1 Setup 
 	HWInit();				//Register initialisieren
 	countf = 0;
 	ctr = 0;

 	mode = MODE;	//Port einlesen
 	dataf = FALSE;
 	SIG1 = 0;
	LED = 0;
	SIG2 = 0;
	ct10 = 10;
	txdready = TRUE;	//RS232 - bertragung beendet
	
/*	
	i = 0;
	txdbuf[i++] = 0x55;
	txdbuf[i++] = 0x41;
	txdbuf[i++] = 13;
	txdbuf[i++] = 10;
	txdbuf[i++] = 0;
	txdidx = 0;
	PIE1bits.TXIE = 1;		//Transmit interrupt ein
*/	
 	while( 1 ) 				//Hauptschleife 
	{
		if( RESET == 0 )	//Reset - Taster
		{
			while( RESET == 0 );
			delay( 5 );				//10ms
			while( RESET == 0 );
			FReset();
		}
		if( dataf == TRUE )
		{
			dataf = FALSE;
			SendData();	
			delay( 0 );		//2 Sekunden warten
			delay( 0 );		//
			delay( 0 );		//
			delay( 0 );		//
			INTCONbits.INT0IF = 0;
			INTCON3bits.INT2IF = 0;
			INTCONbits.INT0IE = 1; 		//external Interrupt 0 wieder ein
			INTCON3bits.INT2IE = 1; 	//external Interrupt 2 wieder ein
		}	
	}
}
//----------------------------------------------------------------------------
// Funktionen:
void HWInit( void )
{
	unsigned char dummy;
	long clock;
	
	ADCON1 = 0x7f;		//0 = analog, 1 = digital
	PORTA = 0;
	TRISA = 0xe1;
	TRISB = 0xfd;	
	ADCON0 = 0;
	//ADCON2 = 0xb2;		//AD-Wandler speed !?
	//ADCON0bits.ADON = 1;
	
//2. UART:
	
	TXSTA = 0x24;
	BAUDCTL = 0x02;
	//SPBRG = 31;				//fr 9600 Baud @ 4,9152 MHZ Quarz
	clock = CLK/16;
	clock /= BAUD;
	//SPBRG = 23;					//fr 9600 Baud @ 3,6864 MHz
	SPBRG = clock - 1;
	RCSTA = 0x90;
	PIR1bits.RCIF = 0;
	//IPR1bits.RCIP = 0;		//low priority int
	PIE1bits.RCIE = 1;		//enable Recive int
	
// 3. PWM:
	//CCP1CON = 0b10001100;
	//PWM1CON = 0x00;
	//T2CON = 0b00010100;	//Timer2 on
	//PR2 = 0xff;
	//CCPR1L = 0x80;

// 4. Timer
	TMR1L = tmr1.b[0];			//Lichtschranken-Funktion
	TMR1H = tmr1.b[1];
	T1CON = 0;				//Vorbereiten aber nicht starten, Vorteiler 1
	T1CONbits.TMR1ON = 0;
	
	TMR3L = 0;			//delay-Funktion
	TMR3H = 0;
	T3CON = 0x30;		//Vorbereiten aber nicht starten, Vorteiler 8
	T3CONbits.TMR3ON = 0;

// 5. Int's:
	RCONbits.IPEN = 0;
	INTCON2bits.INTEDG0 = 0;	//fallende Flake
	INTCON2bits.INTEDG2 = 0;
	INTCONbits.INT0IE = 1; 		//enable external Interrupt 0
	INTCON3bits.INT2IE = 1; 	//enable external Interrupt 2
	PIE1bits.TMR1IE = 1;		//Timer 1 Int enable
	
	PIR1bits.TMR1IF = 0;
	PIE1bits.TMR1IE = 1;	
	
	PIE1bits.TXIE = 0;		//disable Transmit interrupt
	INTCONbits.GIEH = 1;	//Global INT ein
	INTCONbits.PEIE = 1;
}
//-----------------------------------------------------------------------
// FReset stoppt Zhler und setzt ctr auf Null
void FReset( void )
{
	T3CONbits.TMR3ON = 0;	//Timer3  aus
	PIE1bits.TXIE = 0;		//Transmit Interrupt aus
	SIG1 = 0;
	LED = 0;
	SIG2 = 0;
	
	countf = 0;
	ctr = 0;	
	txdready = TRUE;
	
	if( MODE == 0 )
		mode = MODE12;		//beide Lichtschranken
	else
		mode = MODE1;		//nur LS1
}
//------------------------------------------------------------------------
// Funktion Lichtsch ist die Hauptfunktion
//
void lichtsch( int ls )
{
	SIG2 = 1;
	if( mode == MODE12 )	//beide Lichtschranken
	{
		if( countf == 0 )	//zhlt er schon ?
		{
			TMR1L = tmr1.b[0];			//Lichtschranken-Funktion
			TMR1H = tmr1.b[1];
			ctr = 0;
			T1CONbits.TMR1ON = 1;
			countf = ls;
			LED = 1;
			SIG1 = 1;
			//SIG2 = 0;
		}
		else if( ls != countf )		//nur stoppen, wenn ls != countf
		{
			T1CONbits.TMR1ON = 0;		//Timer wieder aus
			INTCONbits.INT0IE = 0; 		//external Interrupt 0 aus
			INTCON3bits.INT2IE = 0; 	//external Interrupt 2 aus
			countf = 0;	
			LED = 0;
			SIG1 = 0;
			//SIG2 = 1;
			dataf = TRUE;
		}
	}
	else	//mode == MODE1
	{
		if( countf == 0 && ls == LS1 )
		{
			TMR1L = tmr1.b[0];			//Lichtschranken-Funktion
			TMR1H = tmr1.b[1];
			ctr = 0;
			T1CONbits.TMR1ON = 1;
			countf = LS1;
			LED = 1;
			SIG1 = 1;
			//SIG2 = 0;
		}
		else if( countf == LS1 && ls == LS1 )
		{
			T1CONbits.TMR1ON = 0;	//Timer wieder aus
			INTCONbits.INT0IE = 0; 		//external Interrupt 0 aus
			INTCON3bits.INT2IE = 0; 	//external Interrupt 2 aus
			countf = 0;	
			LED = 0;
			SIG1 = 0;
			//SIG2 = 1;
			dataf = TRUE;
		}
	}
	SIG2 = 0;
}
//------------------------------------------------------------------------
//delay, CLK/4,:8:256 = 1,66*1ms bei 4,9152 MHz, ohne Interrupt
//delay, CLK/4,:8:256 = 2,22*1ms bei 3,6864 MHz, ohne Interrupt
//
void delay( unsigned char ms )
{
	TMR3L = 0;
	TMR3H =  0-ms;//(0xff ^ ms);	//delay = ms*1,66 oder 2,22 ms
	
	PIR2bits.TMR3IF = 0;
	T3CONbits.TMR3ON = 1;	//Timer start
	while( PIR2bits.TMR3IF == 0 );
	T3CONbits.TMR3ON = 0;	//Timer wieder aus.	
}

//--------------------------------------------------------------------------
// SendChar wird von der TX-ISR aufgerufen
// Sendchar schreibt das nchste byte ins TXREG und berprft ob das Senden 
// beendet ist. Wenn ja, wird der TXD Interrupt ausgeschaltet.
void SendChar( void )
{
	TXREG = txdbuf[txdidx++];
	if( txdbuf[txdidx] == 0 )
	{
		PIE1bits.TXIE = 0;		//Transmit interrupt aus
		txdready = TRUE;
	}	
}
//-------------------------------------------------------------------
// SendByte sendet ein beliebiges Zeichen
void	SendByte( char b )
{
	TXREG = b;
}
//-------------------------------------------------------------------
// SendData sendet Timer inhalt
void	SendData( void )
{
	int 	sctr, i, end;
	
	end = 0;
	sctr = ctr;	
			
	itoa( ctr, txdbuf );
	for( i = 0; i < TXDLEN && end == 0; i++ )
	{
		if( txdbuf[i] == 0 )
			end = 1;
	}
	txdbuf[i-1] = 10;
	txdbuf[i] = '\0';
	//sprintf( txdbuf, "%d\n", sctr );
	txdidx = 0;
	txdready = FALSE;
	PIE1bits.TXIE = 1;		//Transmit interrupt ein
}
//----------------------------------------------------------------------------
// end
